package de.ozgcloud.eingang.semantik.formbased;

import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.api.InstanceOfAssertFactories.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import de.ozgcloud.eingang.semantik.enginebased.ServiceKontoBuildHelper;

class DFoerdermittelFormBasedMapperTest {

	@Spy
	@InjectMocks
	private DFoerdermittelFormBasedMapper mapper;

	@Mock
	private ServiceKontoBuildHelper serviceKontoHelper;

	@Nested
	class TestIsResponsible {

		@Test
		void shouldBeTrueWithFachnachricht() {
			var responsible = mapper.isResponsible(createWithFachnachricht());

			assertThat(responsible).isTrue();
		}

		@Test
		void shouldBeFalseForOuther() {
			var responsible = mapper.isResponsible(FormDataTestFactory.create());

			assertThat(responsible).isFalse();
		}
	}

	@Nested
	class TestParseFachnachricht {

		@Nested
		class ExtractData {
			@Test
			void shouldHaveFormData() {
				var result = mapper.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).isNotNull();
			}

			@Test
			void shouldHavePages() {
				var result = mapper.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsKey("Pages");
			}

			@Test
			void shouldHaveInboxRef() {
				var result = mapper.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsEntry("InboxReference", "sh/sh/4dd01647-b9d9-4775-1b50-08da3d83800a");
			}
		}

		@Nested
		class HandleFachnachrichtData {

			private Map<String, Object> extracted = Map.of("name", "Theo");

			@Test
			void shouldCallExtractData() {
				doReturn(extracted).when(mapper).extractFormDataFormXML(any());

				mapper.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				verify(mapper).extractFormDataFormXML(notNull());
			}

			@Test
			void shouldAddMap() {
				doReturn(extracted).when(mapper).extractFormDataFormXML(any());

				var result = mapper.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).containsEntry("Fachnachricht", extracted);
			}

			@Test
			void shouldIgnoreEmptyData() {
				doReturn(Collections.emptyMap()).when(mapper).extractFormDataFormXML(any());

				var result = mapper.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).doesNotContainKey("Fachnachricht");
			}
		}
	}

	FormData createWithFachnachricht() {

		return FormData.builder()
				.formData(Collections.emptyMap())
				.representation(createFachnachrichtFile()).build();
	}

	IncomingFile createFachnachrichtFile() {
		var tmpFile = TempFileUtils.writeTmpFile(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

		return IncomingFile.builder()
				.file(tmpFile)
				.contentType("application/xml")
				.size(1283)
				.name("Beispieldatensatz_Fachnachricht.xml")
				.build();
	}

	@Nested
	class TestProcessFachnachricht {

		@Captor
		private ArgumentCaptor<Map<String, Object>> fachnachrichtCaptor;

		@Test
		void shouldCallAddServiceKonto() {
			var formData = DFoerdermittelFormDataTestFactory.create();

			mapper.processFachnachricht(formData);

			verify(mapper).addServiceKonto(same(formData), fachnachrichtCaptor.capture());
			assertThat(fachnachrichtCaptor.getValue()).containsAllEntriesOf(DFoerdermittelFormDataTestFactory.createFachnachrichtMap());
		}

		@Test
		void shouldCallAddOrganisationsEinheitId() {
			var extened = DFoerdermittelFormDataTestFactory.create();
			doReturn(extened).when(mapper).addServiceKonto(any(), any());

			mapper.processFachnachricht(DFoerdermittelFormDataTestFactory.create());

			verify(mapper).addOrganisationsEinheitId(same(extened), notNull());
		}
	}

	@Nested
	class TestAddServiceKonto {

		@BeforeEach
		void init() {
			when(serviceKontoHelper.buildOsiServiceKonto(any())).thenReturn(ServiceKontoTestFactory.create());
		}

		@Test
		void shouldHaveServiceKonto() {
			var formData = mapper.addServiceKonto(DFoerdermittelFormDataTestFactory.create(),
					DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			assertThat(formData.getHeader().getServiceKonto().getPostfachAddresses().get(0).getIdentifier())
					.asInstanceOf(type(StringBasedIdentifier.class)).extracting(StringBasedIdentifier::getPostfachId)
					.isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldRemovePrefix() {
			mapper.addServiceKonto(DFoerdermittelFormDataTestFactory.create(), DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			verify(serviceKontoHelper).buildOsiServiceKonto(DFoerdermittelFormDataTestFactory.POSTFACH_ID);
		}
	}

	@Nested
	class TestExtractPrefix {
		@Test
		void shouldRemoveAllBeforeLastSlash() {
			var result = mapper.extractPrefix("bla/bla/bla/12345");

			assertThat(result).isEqualTo("12345");
		}

		@Test
		void shouldBeFineWithoutSlash() {
			var result = mapper.extractPrefix("12345");

			assertThat(result).isEqualTo("12345");
		}
	}

	@Nested
	class TestAddOrganisationsEinheitId {
		@Test
		void shouldHaveOrganisationsEinheitId() {
			var formData = mapper.addOrganisationsEinheitId(DFoerdermittelFormDataTestFactory.create(),
					DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			assertThat(formData.getZustaendigeStelle()).isNotNull().extracting(ZustaendigeStelle::getOrganisationseinheitenId)
					.isEqualTo(DFoerdermittelFormDataTestFactory.ORGANISATIONS_EINHEIT_ID);
		}
	}
}
